!=======================================================================
!
!     this subroutine sets the callback functions 
!     for IPOPT.
!
!=======================================================================

      module IPOPT_wrapper


      use iso_c_binding      

      use parameter_module
      use econ_module

      implicit none
    

   
      contains        



!=======================================================================
!
!     initial value function
!
!=======================================================================
      
      subroutine EV_F(N,X,NEW_X,F,IDAT,DAT,IERR)

      implicit none
            
      integer :: N, NEW_X
      real(prec)   :: F 
      
      real(prec), target :: X(N)

      real(prec)   :: DAT(dim_sim)
      integer :: IDAT(*)
      integer :: IERR

      real(prec), dimension(dim_model) :: grid_pt_box !coords one single grid point on a cubic box            

      integer ::  which_state
                        
!----------------------------------------------------------------------   

!.....current state


      call objective_init( N, X, &
          & F, idat, DAT )

            
      IERR = 0
    
      return
!-----------------------------------------------------------------------      
      end subroutine EV_F
!-----------------------------------------------------------------------         
      
      
!=======================================================================
!
!     value function objective during VFI
!
!=======================================================================
      subroutine EV_F_ITER(N,X,NEW_X,F,IDAT,DAT,IERR)

      implicit none
            
      integer :: N, NEW_X
      real(prec)   :: F 
      real(prec), target :: X(N)

      real(prec)   :: DAT(dim_sim)
      integer :: IDAT(*)
      integer :: IERR

      real(prec), dimension(dim_model) :: grid_pt_box !coords one single grid point on a cubic box            
                      
      integer :: which_state


!----------------------------------------------------------------------   
      

!.....current state


      call objective_iter( N, X, &
         & F, idat, DAT )


           
      IERR = 0
      
      return
!-----------------------------------------------------------------------      
      end subroutine EV_F_ITER
!-----------------------------------------------------------------------        


!=============================================================================
! 
!     Computation of gradient of objective function
! 
!=============================================================================

      subroutine EV_GRAD_F(N, X, NEW_X, GRAD, IDAT, DAT, IERR)
      
      implicit none
      
      integer :: N, NEW_X
      real(prec)   :: GRAD(N), X(N)
      real(prec),dimension(dim_sim) :: DAT
      integer :: IDAT(no_nz_obj+1)
      integer :: IERR
      integer :: indxoffs

      real(prec) :: h, fx1, fx2, xAdj(N)

      integer :: ixN,ixC

      integer   ::  AVAR_OBJ(no_nz_obj)

!-----------------------------------------------------------------------     
      
      GRAD(:) = 0.0  !init. gradient of objective fct.

      h = 1.d-4

      AVAR_OBJ(:) = IDAT(2:no_nz_obj+1)

      call EV_F(N, X, NEW_X, fx1, IDAT, DAT, IERR)

      do ixC=1,no_nz_obj
                ixN = AVAR_OBJ(ixC)  !only compute the derviative where it is potentially nonzero
		xAdj=x
		xAdj(ixN)=x(ixN)+h
		call EV_F(N, xAdj, NEW_X, fx2, IDAT, DAT, IERR)          
		xAdj(ixN)=x(ixN)-h
		call EV_F(N, xAdj, NEW_X, fx1, IDAT, DAT, IERR)
		GRAD(ixN)=(fx2-fx1)/(2.0*h)
      enddo

      IERR = 0
    ! write(*,*) 'x = ', x
    ! write(*,*) 'grad = ', grad
    ! write(*,*) 'return EV_GRAD_F'
      return
!-----------------------------------------------------------------------     
      end subroutine EV_GRAD_F
!-----------------------------------------------------------------------     
 
 
 !=============================================================================
 ! 
 !     Computation of gradient of objective function for iteration
 ! 
 !=============================================================================
 ! 
      subroutine EV_GRAD_F_ITER(N, X, NEW_X, GRAD, IDAT, DAT, IERR)
      
      implicit none
      
      integer :: N, NEW_X
      real(prec)   :: GRAD(N), X(N)
      real(prec),dimension(dim_sim) :: DAT
      integer :: IDAT(*)
      integer :: IERR
      integer :: indxoffs

      real(prec) :: h, fx1, fx2, xAdj(N)
      integer :: ixN,ixC

      integer   ::  AVAR_OBJ(no_nz_obj)

!-----------------------------------------------------------------------     
      
      GRAD(:) = 0.0  !init. gradient of objective fct.

      AVAR_OBJ(:) = IDAT(2:no_nz_obj+1)

      h = 1.d-4

      call EV_F_ITER(N, X, NEW_X, fx1, IDAT, DAT, IERR)


      do ixC=1,no_nz_obj
                ixN = AVAR_OBJ(ixC)  !only compute the derviative where it is potentially nonzero
		xAdj=x
		xAdj(ixN)=x(ixN)+h
		call EV_F_ITER(N, xAdj, NEW_X, fx2, IDAT, DAT, IERR)          
		xAdj(ixN)=x(ixN)-h
		call EV_F_ITER(N, xAdj, NEW_X, fx1, IDAT, DAT, IERR)
		GRAD(ixN)=(fx2-fx1)/(2.0*h)
      enddo

      IERR = 0

      return
!-----------------------------------------------------------------------     
      end subroutine EV_GRAD_F_ITER
!-----------------------------------------------------------------------   

!=============================================================================
! 
!     Computation of equality constraints 
! 
!=============================================================================

      subroutine EV_G(N, X, NEW_X, M, G, IDAT, DAT, IERR)
        
      implicit none
      
      integer :: N, NEW_X, M
      real(prec)   :: GRAD(N)
      real(prec), target   :: X(N)
      real(prec),dimension(dim_sim) :: DAT
      integer :: IDAT(*)
      integer :: IERR
      real(prec)              :: G(M)       

      integer                 ::  indxit

      real(prec), pointer :: U1, U2, futurutil1P1, futurutil1P2, &
                    & futurutil2P1, futurutil2P2

      real :: w1,w2

      real(prec),dimension(dim_sim)  ::  grid_pt_box                
                
      
!----------------------------------------------------------------------   


      do indxit =1,dim_sim      
         grid_pt_box(indxit) = DAT(indxit)
      enddo

!.....mapping the x to named pointers
 

      call constraints(N, X, M, G, IDAT, DAT)



    
      IERR = 0

      return
      
!-----------------------------------------------------------------------      
      end subroutine EV_G
!-----------------------------------------------------------------------      


   
! 
! 

!=============================================================================
! 
!                 Computation of Jacobian of equality constraints
! 
!=============================================================================

      subroutine EV_JAC_G(TASK, N, X, NEW_X, M, NZ, ACON, AVAR, A, &
     &     IDAT, DAT, IERR)

      integer :: TASK, NEW_X, NZ
      integer :: M, N
      real(prec), target :: X(N)
      real(prec) :: A(NZ)
      integer :: ACON(NZ), AVAR(NZ)
      integer :: i, j, k
      real(prec),dimension(dim_sim) :: DAT
      integer :: IDAT(*)
      integer :: IERR

      integer :: ixM, ixN
!-----------------------------------------------------------------------      


!-----------------------------------------------------------------------      
!    !jacobian matrix structure      
      if( TASK.eq.0 ) then


        call sparsity_pattern(ACON, AVAR, NZ)



      else

!-----------------------------------------------------------------------
!.....finite differences 

          call jac_constraints(N, X, M, NZ, A, &
             &     IDAT, DAT)


      endif

      IERR = 0


!-----------------------------------------------------------------------      
      return
!-----------------------------------------------------------------------      
      end subroutine EV_JAC_G
      
!  =============================================================================      
      

      
!  =============================================================================
! 
!                 Computation of Hessian of Lagrangian
! 
!  =============================================================================

      subroutine EV_HESS(TASK, N, X, NEW_X, OBJFACT, M, LAM, NEW_LAM, &
     &     NNZH, IRNH, ICNH, HESS, IDAT, DAT, IERR)
     
      implicit none
     
      integer :: TASK, N, NEW_X, M, NEW_LAM, NNZH, i
      real(prec)   ::  OBJFACT, LAM(M), HESS(NNZH)
      real(prec), target      :: X(N)

      integer :: IRNH(NNZH), ICNH(NNZH)

      real(prec) :: z_shock
      integer :: state_now_in, current_state
!      real(prec)   :: DAT(*)
      real(prec)   :: DAT(dim_sim)
      integer :: IDAT(*)
      integer :: IERR
            
      

      IERR = 1
!-----------------------------------------------------------------------            
      return
!-----------------------------------------------------------------------      
      end subroutine EV_HESS
!-----------------------------------------------------------------------


!=======================================================================
! 
!                    Callback method called once per iteration
! 
!=======================================================================

      subroutine ITER_CB_N(ALG_MODE, ITER_COUNT,OBJVAL, INF_PR, INF_DU, &
     &     MU, DNORM, REGU_SIZE, ALPHA_DU, ALPHA_PR, LS_TRIAL, IDAT, &
     &     DAT, ISTOP)
     
      implicit none
     
      integer :: ALG_MODE, ITER_COUNT, LS_TRIAL
      real(prec)   :: OBJVAL, INF_PR, INF_DU, MU, DNORM, REGU_SIZE
      real(prec)   :: ALPHA_DU, ALPHA_PR
      real(prec)   :: DAT(dim_sim)
      integer :: IDAT(*)
      integer :: ISTOP

!     You can put some output here

      !write(*,*) 'Testing callback function in iteration ', ITER_COUNT

!     And set ISTOP to 1 if you want Ipopt to stop now.  Below is just a
!     simple example.

      if (INF_PR.le.1D-04) ISTOP = 1
!-----------------------------------------------------------------------      
      return
!-----------------------------------------------------------------------            
      end subroutine ITER_CB_N
!-----------------------------------------------------------------------            
      
 



     
      
!-----------------------------------------------------------------------   
      
      end module IPOPT_wrapper
      
!***********************************************************************


