!=======================================================================
!
!     this module sets callback functions 
!     for SNOPT.
!
!=======================================================================

      module SNOPT_wrapper

      use parameter_module
      use econ_module

      implicit none
    

   
      contains        



!=======================================================================
!
!     initial value function
!
!=======================================================================

 
!     ==================================================================
!     Computes the nonlinear objective and constraint terms for a
!     modified version of HS47.
!     ==================================================================      
      subroutine EV_F( Status, n, x, &
     &     needF, nF, F, &
     &     needG, lenG, G, &
     &     cu, lencu, iu, leniu, ru, lenru )

      implicit none
      integer ::  Status, needF, needG, nF, n, lenG, &
     &     lencu, leniu, lenru, &
     &     iu(leniu)
      real(8) ::   F(nF), G(lenG), x(n), &
     &     ru(lenru)
      character(LEN=8) ::  cu(lencu)

!.....Auxiliariy variables
      integer ::  Obj  !location of objective function
      integer :: indx,indx_offset,indxh
      real(prec) :: util, x_tmp(n)

      real(prec), dimension(nvars)    :: xhm,xhp !finite difference variables
      real(prec) :: h
      real(prec) :: fvalhm,fvalhp

      real(prec) :: tmp_vec(no_nz_jac)
      real(prec) :: tmp_convec(nconstr)
      real(prec),dimension(dim_sim) :: DAT
      integer,dimension(1+no_nz_obj) :: IDAT

      integer   ::  AVAR_OBJ(no_nz_obj)


!----------------------------------------------------------------------   

      Obj    = nconstr+1                ! index of objective row

      IDAT(:) = iu(:)
      DAT(:) = ru(:)


      x_tmp(:) = x(:)   !making sure the right real type is passed to non snopt functions
      if (needF .gt. 0) then   !provide objective and constraints
!.....call current functions

              call objective_init( n, x_tmp, &
              & util, IDAT, dat )

	      F(Obj) = util
		    
	      call constraints(N, x_tmp, nF-1,tmp_convec , IDAT, DAT)
              do indx=1,nconstr
                  F(indx) = tmp_convec(indx)
              enddo

      endif

      if (needG .gt. 0) then

!.........constraint derivative
          call jac_constraints(nvars, x_tmp, nconstr, no_nz_jac, tmp_vec, &
          &     IDAT, DAT)

          do indx=1,no_nz_jac
             G(indx) = tmp_vec(indx)
          enddo

!.........objective derivative
!.........computing derivative with finite difference

          AVAR_OBJ(:) = iu(2:no_nz_obj+1)
          indx_offset = no_nz_jac

          if(indx_offset + no_nz_obj .ne. lenG) then
              Write(*,*) &
              & "ERROR in objective: derivative numbers wrong. indx_offset + no_integnodes*dim_model",&
              & indx_offset + no_integnodes*dim_model, "lenG",lenG
          endif

          h = 1d-3

          do indx =1,no_nz_obj
              xhm(:) = x_tmp(:)
              xhp(:) = x_tmp(:)
              indxh = AVAR_OBJ(indx) !only compute the derivate at the nonzero entries
              xhm(indxh) = x(indxh) - h
              
              call objective_init( n, xhm, &
              & fvalhm, idat, dat )

              xhp(indxh) = x(indxh) + h
              call objective_init( n, xhp, &
              & fvalhp, idat, dat )
              G(indx_offset + indx) = (fvalhp - fvalhm)/(2.0*h)
          enddo


      endif


      return
!-----------------------------------------------------------------------      
      end subroutine EV_F
!-----------------------------------------------------------------------       


     
!=======================================================================
!
!     value function objective during VFI
!
!=======================================================================
      subroutine EV_F_ITER( Status, n, x, &
     &     needF, nF, F, &
     &     needG, lenG, G, &
     &     cu, lencu, iu, leniu, ru, lenru )

      implicit none
      integer :: Status, needF, needG, nF, n, lenG, &
     &     lencu, leniu, lenru, &
     &     iu(leniu)
      real(8) :: F(nF), G(lenG), x(n), &
     &     ru(lenru)
      character(LEN=8) :: cu(lencu)

      integer ::  Obj  !location of objective function
      integer :: indx,indxoffs,indxh
      real(prec) :: util, x_tmp(n)

      real(prec), dimension(nvars)    :: xhm,xhp !finite difference variables
      real(prec) :: h
      real(prec) :: fvalhm,fvalhp
             
                        
!.....Fortran - C interface 
      integer :: nst, which_GP, which_state

      real(prec) :: query_pt(dim_sim),mu1,stdev
!.....Auxiliariy variables
      real(prec) :: tmp_vec(no_nz_jac)
      real(prec) :: tmp_convec(nconstr)
      real(prec),dimension(dim_sim) :: DAT
      integer,dimension(1+no_nz_obj) :: IDAT

      integer   ::  AVAR_OBJ(no_nz_obj),indx_offset
 
!----------------------------------------------------------------------   
  
      Obj    = nconstr+1                ! index of objective row    

!.....current state
      which_state = iu(1)
      IDAT(:) = iu(:)
      DAT(:) = ru(:)

      do indx=1,nvars
          x_tmp(indx) = x(indx)   !making sure the right real type is passed to non snopt functions
      enddo
      if (needF .gt. 0) then   !provide objective and constraints
!.....call current functions

              call objective_iter( n, x_tmp, &
              & util, idat, dat )

!.....output      
	      F(Obj) = util
		    
	      call constraints(N, x_tmp, nF-1,tmp_convec , IDAT, DAT)
              do indx=1,nconstr
                  F(indx) = tmp_convec(indx)
              enddo

      endif  

      if (needG .gt. 0) then  

          call jac_constraints(nvars, x_tmp, nconstr,no_nz_jac, &
          &  tmp_vec, idat, dat)
          G(1:no_nz_jac) = tmp_vec(:)

!.........objective derivative
!.........computing derivative with finite difference

          AVAR_OBJ(:) = iu(2:no_nz_obj+1)
          indx_offset = no_nz_jac

          if(indx_offset + no_nz_obj .ne. lenG) then
              Write(*,*) &
              & "ERROR in objective: derivative numbers wrong. indx_offset + no_integnodes*dim_model",&
              & indx_offset + no_integnodes*dim_model, "lenG",lenG
          endif

          h = 1d-3

          do indx =1,no_nz_obj
              xhm(:) = x_tmp(:)
              xhp(:) = x_tmp(:)
              indxh = AVAR_OBJ(indx) !only compute the derivate at the nonzero entries
              xhm(indxh) = x(indxh) - h
              
              call objective_iter( n, xhm, &
              & fvalhm, idat, dat )

              xhp(indxh) = x(indxh) + h
              call objective_iter( n, xhp, &
              & fvalhp, idat, dat )
              G(indx_offset + indx) = (fvalhp - fvalhm)/(2.0*h)
          enddo
 
      endif


      return
!-----------------------------------------------------------------------      
      end subroutine EV_F_ITER
!-----------------------------------------------------------------------        


     
!-----------------------------------------------------------------------   
      
      end module SNOPT_wrapper
      
!***********************************************************************


