!=======================================================================
!
!     subroutine nonlinear_solver_sim_snopt 
!
!=======================================================================


      
      subroutine nonlinear_solver_sim_snopt(which_state, grid_pt_box, &
                          & start_pt,use_warmstart, output_solver, return_status)
  

      use, intrinsic :: iso_fortran_env, only : stdin=>input_unit, &
                                          stdout=>output_unit, &
                                          stderr=>error_unit

      USE, INTRINSIC :: ISO_C_BINDING
      
      use econ_module


      use SNOPT_wrapper, only: EV_F_ITER
                  
      use parameter_module
                  
      
  
      implicit none
      
      
!.....input variables      
      integer, intent(in) :: which_state!in which state are we 
      integer, intent(in) :: use_warmstart!in which state are we 
      real(prec), dimension(nvars), intent(in) :: start_pt  !in which state are we 
      real(prec), dimension(dim_sim), intent(in) :: grid_pt_box !coords one single grid point on a cubic box

!----------------------------------------------------------------------      
!     IPOPT PARAMETERS below:
!----------------------------------------------------------------------
!.....Size of the problem (number of variables and equality constraints)       
      integer :: M != dim_sim + 1, !number of constraints
    

!.....Private data for evaluation routines
!     This could be used to pass real(prec) ::  and integer arrays untouched
!     to the evaluation subroutines EVAL_*
      real(prec) :: DAT(dim_sim)
      integer    :: IDAT(1+no_nz_obj)

!-----------------------------------------------------------------------        
!.....SNOPT variables
!-----------------------------------------------------------------------  

      integer,  parameter :: maxF   = nconstr+1, maxn = nvars, &
     &     lenA   = 1, & !no_nz_lin, &
     &     lenG   = no_nz_jac + no_nz_obj, &  !no_nz_nl_wObj, &
     &     nxname =  1, nFname =   1 
      integer :: iAfun(lenA), jAvar(lenA), iGfun(lenG), jGvar(lenG), &
     &     xstate(maxn), Fstate(maxF)
      CHARACTER(LEN=8) ::  Probl, xnames(nxname), Fnames(nFname)
      real(8) :: ObjAdd, sInf,&
     &     A(lenA), Flow(maxF), Fupp(maxF), F(maxF), &
     &     xlow(maxn), xupp(maxn), x(maxn), Fmul(maxF), xmul(maxn)
      integer ::  DerOpt, neA, neG, ObjRow, INFO, iPrt, iPrint, iSpecs, &
     &     iSum, iSumm, Mjrlim, mincw, miniw, minrw, nF, n, nInf, &
     &     nOut, nS, Obj

      integer :: lunit
      CHARACTER(LEN=20) :: lfile

!     ------------------------------------------------------------------
!     SNOPT workspace

      integer, parameter :: lenrw = 800*(nvars + nconstr) + 1000
      real(8) ::  rw(lenrw)
      integer,  parameter :: leniw = 200*(nvars + nconstr) + 1000
      integer  ::       iw(leniw)
      integer , parameter :: lencw =   500 + nvars + nconstr
      CHARACTER(LEN=8)    ::  cw(lencw)

!     ------------------------------------------------------------------
!     SNOPT variables used for passing information

      integer ,parameter :: lenru = dim_sim
      real(8)  :: ru(lenru)
      integer,parameter ::  leniu = 1+no_nz_obj
      integer  ::   iu(leniu)
      integer   ,parameter ::  lencu = 1 
      character(LEN=8)  ::  cu(lencu)

!     ------------------------------------------------------------------
      integer ,parameter :: Cold   = 0, Basis  = 1, Warm  = 2
!     ------------------------------------------------------------------


!-----------------------------------------------------------------------        
!.....local variables                
!-----------------------------------------------------------------------  

      real(8), dimension(nrestarts,nvars)  :: X_start 
      real(prec), dimension(nrestarts,nvars)  :: X_start_real 
      integer :: i,j,max_indx,indx_offset  !index
      real(prec) :: rand_num, X_real(nvars)
      real(prec), allocatable ::  results_allrestarts(:,:)  !stores all policy values, objective value and return status for each restart
      integer, dimension(33) :: seed



!.....state space bounds
      integer   ::  ACON(no_nz_jac),AVAR(no_nz_jac),AVAR_OBJ(no_nz_obj)

!-----------------------------------------------------------------------        
!.....debug variables                
!-----------------------------------------------------------------------  

!      real(prec), dimension(:,:), allocatable   :: test_mat
!      real(prec), dimension(:), allocatable     :: test_likel_vec
!      real(prec), dimension(dim_sim)          :: temp_vec
!      integer                                   :: no_pts

!-----------------------------------------------------------------------        
!.....output variables                
!----------------------------------------------------------------------- 
      
      real(prec), intent(out) :: output_solver(2+nvars)
      logical, intent(out) :: return_status !0 if converged -1 otherwise
 

!==================================================================
!==== Main function ==============================================
!==================================================================


!.....define problem size and allocate corresponding memory:
      N = nvars    !number of vars.     
      M = nconstr    !number of constraints
      neG = lenG  !entries of Jacobian matrix of contraints
      neA = 0 !lenA   
      A(:) = 0 !no entries in linear constraints
      iw(:) = 0
      rw(:)=0.
      cw(:) = "" 
        

      allocate(results_allrestarts(nrestarts,N+3))   !stores all policy values, objective value and return status

!     ------------------------------------------------------------------
!     Specify some of the SNOPT files.
!     iSpecs  is the Specs file   (0 if none).
!     iPrint  is the Print file   (0 if none).
!     iSumm   is the Summary file (0 if none).

      iSpecs =  0!6

      if (debug .le. 4) then
         iSumm  =  0!6
         iPrint = 0!6
      else
         iSumm  =  6
         iPrint = 15
         lunit = iPrint
         lfile = 'snopt_print.out'
         open( lunit, file=lfile )
         write(iPrint,*) "============================================================"
         write(iPrint,*) "current Halton point is",grid_pt_box(:)
         write(iPrint,*) "============================================================"
      endif

!        Unix and DOS systems.  Open the Specs and print files.

      iPrt   = iPrint
      iSum   = iSumm

!     ------------------------------------------------------------------
!     First,  snInit MUST be called to initialize optional parameters
!     to their default values.
!     ------------------------------------------------------------------
      call snInit( iPrint, iSumm, cw, lencw, iw, leniw, rw, lenrw )



!     ------------------------------------------------------------------
!     Give the problem a name.

      Probl   = 'P'  

      nF     = nconstr+1        ! constraints and objective
      Obj    = nconstr+1                ! index of objective row
      ObjRow = nconstr+1        ! Indx where objective is stored


      xstate(:) =  0    !is a set of initial states for each x  (0,1,2,3,4,5).

      Fmul(:) = 0       !is a set of initial values for the dual variables.


!.....bounds on constraints and variables

      call bounds(n,xlow, xupp, m, flow(1:M), fupp(1:M), bigbound, &
      &                                          IDAT, DAT )

      Flow(nconstr+1) = -bigbound   !bounds on the objective
      Fupp(nconstr+1) = bigbound


!.........................................................................
!.....sparsity pattern


       call sparsity_pattern(ACON, AVAR, no_nz_jac)
       do j=1,no_nz_jac   
            iGfun(j) = ACON(j)
            jGvar(j) = AVAR(j)
       enddo

       call sparsity_pattern_obj(AVAR_OBJ, no_nz_obj)
       do j=1,no_nz_obj
           iGfun(no_nz_jac+j) = ObjRow
           jGvar(no_nz_jac+j) = AVAR_OBJ(j)
       enddo

!-----------------------------------------------------------------------       
!      passing info to IPOPT or SNOPT

 
      do j =1,dim_model
         DAT(j) = grid_pt_box(j)
      enddo
      
      ru(:) = DAT(:)

    

      iu(1) = which_state
      iu(2:1+no_nz_obj) = AVAR_OBJ(:)
      IDAT(:) = iu(:)

!...............................................................................
!.....bounds on constraints and variables

      call bounds(n,xlow, xupp, m, flow(1:M), fupp(1:M), bigbound, &
      &                                          IDAT, DAT )

      Flow(nconstr+1) = -bigbound   !bounds on the objective
      Fupp(nconstr+1) = bigbound




!-----------------------------------------------------------------------       
!     compute starting points

      X_start(:,:) = X_start_real(:,:)
      do j=1,nrestarts
           seed(:) = (j*1234+4356)
           call random_seed (PUT = seed) 
           do i = 1,nvars
               call random_number(rand_num)
               X_start(j,i) = xlow(i) + (xupp(i)-xlow(i))*rand_num
           enddo
           X_start(j,11:N) = X_start(j,11:N)/1000.
           call solve_feas(grid_pt_box, X_start(j,i), nvars, DAT, IDAT) !giving a better guess than random
      enddo

      x(:) = X_start(2,:)


!     ------------------------------------------------------------------
!     Warn snOptA that userf does not set the derivatives.
!     ------------------------------------------------------------------

      if(debug .le. 2) then
	      call snSeti( 'Verify level', -1, iPrt, iSum, INFO,&
	     &     cw, lencw, iw, leniw, rw, lenrw )
	      call snSet( 'Major Print level 000000', iPrt, iSum, INFO,&
	     &      cw, lencw, iw, leniw, rw, lenrw )

	      call snSeti(  'Minor print level ', 0,  iPrt, iSum, INFO,&
	     &     cw, lencw,  iw, leniw, rw, lenrw )
      else
	      call snSeti( 'Verify level', 3, iPrt, iSum, INFO,&
	     &     cw, lencw, iw, leniw, rw, lenrw )
	      call snSet( 'Major Print level 000011', iPrt, iSum, INFO,&
	     &      cw, lencw, iw, leniw, rw, lenrw )

	      call snSet( 'Solution   yes', iPrt, iSum, INFO,&
	     &      cw, lencw, iw, leniw, rw, lenrw )

	      call snSeti(  'Minor print level', 1,  iPrt, iSum, INFO,&
	     &      cw, lencw, iw, leniw, rw, lenrw )

      endif

      DerOpt = 1

      call snSeti( 'Derivative option', DerOpt, iPrt, iSum, INFO,&
     &    cw, lencw, iw, leniw, rw, lenrw )

      MjrLim = 500
      call snSeti( 'Major iterations limit', MjrLim, iPrt, iSum, INFO,&
     &      cw, lencw, iw, leniw, rw, lenrw )
      call snSeti( 'Minor iterations limit', MjrLim/2, iPrt, iSum, INFO,&
     &      cw, lencw, iw, leniw, rw, lenrw )

       call snSet('Hessian full memory' , iPrt, iSum, INFO,&
     &     cw, lencw, iw, leniw, rw, lenrw )

      call snSet('Maximize' , iPrt, iSum, INFO,&
     &     cw, lencw, iw, leniw, rw, lenrw )




!      call snSetr('LU factor tolerance',1.0 , iPrt, iSum, INFO, &
!      &      cw, lencw, iw, leniw, rw, lenrw )
!      call snSetr('LU update tolerance',1.0 , iPrt, iSum, INFO, &
!      &      cw, lencw, iw, leniw, rw, lenrw )

!      call snSet('Nonderivative linesearch' , iPrt, iSum, INFO,&
!     &     cw, lencw, iw, leniw, rw, lenrw )

!      call snSetr('Minor feasibility tolerance',eps_error , iPrt, iSum, INFO, &
!      &      cw, lencw, iw, leniw, rw, lenrw )
      call snSetr('Function precision',1e-6 , iPrt, iSum, INFO, &
      &       cw, lencw, iw, leniw, rw, lenrw )
      call snSetr('Major feasibility tolerance',1e-6 , iPrt, iSum, INFO, &
      &       cw, lencw, iw, leniw, rw, lenrw )
      call snSetr('Major optimality tolerance',1e-3 , iPrt, iSum, INFO, &
      &      cw, lencw, iw, leniw, rw, lenrw )
      !call snSetr('Infinite Bound size',bigbound , iPrt, iSum, INFO, &
      !&      cw, lencw, iw, leniw, rw, lenrw )
!      call snSetr('Unbounded objective',-10.*min_GP(which_state) , iPrt, iSum, INFO, &
!      &      cw, lencw, iw, leniw, rw, lenrw )


!     ------------------------------------------------------------------
!     Go for it, using a Cold start (Start = 0).
!     ------------------------------------------------------------------

      ObjAdd = 0.  !constant added to objective


!.....solve nonlinear system of equations.
      !write(*,*) "hello from thread ", rank

      !$OMP PARALLEL DO schedule(dynamic) DEFAULT(FirstPrivate),shared(X_start), &
      !$OMP SHARED(results_allrestarts)
      do i =1,nrestarts

         call snOptA( Cold, nF, n, nxname, nFname,&
            &     ObjAdd, ObjRow, Probl, EV_F_ITER,&
            &     iAfun, jAvar, lenA, neA, A,&
            &     iGfun, jGvar, lenG, neG,&
            &     xlow, xupp, xnames, Flow, Fupp, Fnames,&
            &     X_start(i,:), xstate, xmul, F, Fstate, Fmul,&
            &     INFO, mincw, miniw, minrw,&
            &     nS, nInf, sInf,&
            &     cu, lencu, iu, leniu, ru, lenru,&
            &     cw, lencw, iw, leniw, rw, lenrw )

         

         if (INFO .eq. 82 .or. INFO .eq. 83 .or. INFO .eq. 84) then
             write(*,*) "ERROR: workspace for SNOPT too small, error code",info
             !exit
         end if


         if((INFO .eq. 1) .or. (INFO .eq. 2) .or. (INFO .eq. 3)) then

             results_allrestarts(i,1) = 0
             results_allrestarts(i,2) = F(ObjRow)
             x_real(:) = X_start(i,:)
             if(debug .ge. 3) then
                 write(*,*) "==================================================="
                 write(*,*) "Intermediate result in nonlinear_solver_iterate"
                 write(*,*) "Current point", grid_pt_box(:)
                 write(*,*) "Objective",results_allrestarts(i,2) 
                 write(*,*) "Optimal solution",X_start(i,:)
                 write(*,*) "==================================================="
             endif
         else
             results_allrestarts(i,1) = -1
             results_allrestarts(i,2) = -1d20
         endif
           
         results_allrestarts(i,4:N+3) = X_start(i,:)

      enddo
      !$OMP END PARALLEL DO  

!-----------------------------------------------------------------------            
!.....Output & refinement criterion:

      output_solver(1) = -1d20 !objective
      output_solver(2:nvars+1) = 0.
      return_status = .false.
      do i =1,nrestarts

          if(  results_allrestarts(i,1) .eq. 0 .and. & 
             & results_allrestarts(i,2) .ge. output_solver(1) ) then
              
              output_solver(1) = results_allrestarts(i,2)
              output_solver(2) = results_allrestarts(i,3)
              output_solver(3:N+2) = results_allrestarts(i,4:N+3)
              return_status = .true.

          endif

      enddo
      if( return_status .eqv. .true. ) then
            if(debug .ge. 2) then
                 write(*,*) "========================================="
                 write(*,*) "CONVERGED in nonlinear_solver_iterate"
                 write(*,*) "Current point", grid_pt_box(:)
                 write(*,*) "Objective",output_solver(1) 
                 write(*,*) "========================================="
            endif
            if(debug .ge. 4) then
                 write(iPrint,*) "========================================="
                 write(iPrint,*) "CONVERGED in nonlinear_solver_iterate"
                 write(iPrint,*) "Current point", grid_pt_box(:)
                 write(iPrint,*) "Objective",output_solver(1) 
                 write(iPrint,*) "========================================="
            endif

      else


            if(debug .ge. 2) then        
               write(*,*) "============================================================"
               write(*,*) 'Nonlinear_solver_iterate did not converge at any point.'
               write(*,*) "Current point", grid_pt_box(:)
               write(*,*) "============================================================"
            endif  
            if(debug .ge. 4) then
                 write(iPrint,*) "========================================="
                 write(iPrint,*) 'Nonlinear_solver_iterate did not converge at any point.'
                 write(iPrint,*) "Current point", grid_pt_box(:)
                 write(iPrint,*) "========================================="
            endif         

      endif
        

! !.....Clean up

            
!-----------------------------------------------------------------------      
     deallocate(results_allrestarts) 
      
!-----------------------------------------------------------------------        
     

     

!=======================================================================

     end subroutine nonlinear_solver_sim_snopt

!***********************************************************************
      
      
