
      subroutine main_fortran()       
!***********************************************************************
!
!     NOTE: Parameters need to be set in parameter_module.f90
!
!***********************************************************************


      use parallelization
      use parameter_module
      use econ_module
      USE TasmanianSG
    
      implicit none

      include 'mpif.h'

      integer       :: step                !iteration step 
      integer(C_INT) :: step_c
      integer(C_INT) :: state_c
      integer       :: DofPerNode          !how many controls

      real(prec), allocatable    :: output_solver(:) !solution to econ problem      
      real(8), allocatable    :: solve_data(:,:)  !matrix of values for SG

      real(8), allocatable    :: solve_data_loc(:)  !matrix for val of SG tmp

      logical, allocatable :: return_status_vec(:)
      logical, allocatable :: return_status_vec_loc(:)


!.....tasmanian grid vars

      double precision, pointer :: points(:,:)
      real(C_DOUBLE) :: transform_up(dim_model),transform_down(dim_model)  !state space bounds used in tranformation
      integer(C_INT) :: level
      integer(C_INT), dimension(n_disc_state) :: No_grid_pts_vec   !Number of grid points for VFI
      integer(C_INT) ::  dim_out       !R^n -> R^m (dim_statespace -> dim_out) for interpolgrids

!.....grid points for passing to solver
      real(prec), dimension(dim_model) :: grid_pt_box !coords one single grid point grid_pt(dim) in [0..1]^d


      character(80)  :: filename = ' '

!.....OMP parallelization

      integer :: num_threads,omp_chunk_size
      integer :: tmp_bound
      integer , external :: OMP_GET_THREAD_NUM, OMP_GET_NUM_THREADS

!......Fortran - C interface 

      character(len=256) :: filename_type1
      character(len=256) :: format_string_step,format_string_st    
      character(len=256) :: folder    
      integer            :: str_counter, nst
      
      integer(C_INT) :: dout !dimension of range of function to be approxmiated (>1 if we want to add the policies)

!.....indices
      integer :: indx_s,indx_pt, counter,indx_s2,indxd, indxsum, indxout
      integer       :: indxi,j      !loop indices


!.....MPI variables
      integer :: no_feas,no_feas_local, indx_ref

      integer :: mpi_ierr, n_process_comm
      integer(C_INT) :: root,rank

      integer :: state_start, state_end
      integer :: indx_start, indx_end
      integer :: mpi_chunk_size
      real(prec)   :: wtime,wtimeomp     !wall time for computation

      integer :: getRoot,getProcessCount,getRank,getComm,isMaster



!........................................................................

      num_threads = OMP_GET_NUM_THREADS() !maximal number of threads possible in the program


!....................................................................... 
!.....basic info on model



      if( proc%isMaster()) then
          write(*,*) '======================================================='
          write(*,*) "                    Start of VFI"
      
!.....set result filename
          if(debug .ge. 2) then
             write(filename,'(a)')'results_timestep.txt'
             write(*,*)'fn=',trim(filename)      
             open (unit = 2, file = filename)  
          end if  
          write(*,*) '======================================================='
      endif

      DofPerNode = nvars + 1 +dim_model  !degrees of freedom per node


!........................................................................

    
!-----------------------------------------------------------------------      
!.....loop over the DP iteration steps      
!-----------------------------------------------------------------------      

      do step = start_iter, end_iter

          step_c = step !to pass to C

!.........................................................................


          if (step > 1) then
!----------------------------------------------------------------------
!.......load GP into memory if we are just starting out, otherwise it is loaded at the end
                if(proc%isMaster()) then
                       write(*,*) &
                 &"============load data of agents ===================="
                endif

!.........generate filename     
                do nst=1,n_disc_state
                   write(format_string_step,'(I0)') step-1  !previous iterations' data
                   write(format_string_st,'(I0)')  nst
                   write(filename_type1, '(a,a)') trim(restart_dir)// &
                   &             'VF_step_'//trim(format_string_step)&
                   &                      //'_'//trim(format_string_st)
                   if(proc%isMaster()) then
                        write(*,*)'fn=',trim(filename_type1)
                   endif
                   call tsgAllocateGrid(grid_old(nst))  !init object
                   if (.not. tsgRead(grid_old(nst), trim(filename_type1) // c_null_char)) then 
                      write(*,*) "File read failed: 1"
                      stop 1
                   end if
                end do

              
                if(proc%isMaster()) then
                     write(*,*) &
                     &    "============ old GPs loaded ================"
                endif


          endif

        
 
!........................................................................
!.....MPI setup

          ! Dedicate MPI process to parallel state
          ! (or do nothing, if parallel states are not enabled).
          ! FIXME: This will not work if we do not have static number of points
          !        Then we require each comm to compute its number of points and
          !        do a sync over all the communicators


!          call proc%dedicate(n_disc_state, no_grid_pts_vec, parstates, &
!               &                                   step .eq. start_iter)

          root = proc%getRoot(proc%getComm())
          n_process_comm = proc%getProcessCount(proc%getComm())
          rank = proc%getRank()

          write(*,*) "Hello from comm", proc%getCommIndex(),"rank",rank, &  
                 &         "with root",root, "is Master",proc%isMaster()

!-----------------------------------------------------------------------    

          if(proc%isMaster()) then
                wtime = MPI_Wtime( )
          endif


!======================================================================

!...... Split the loop below across MPI groups, if parallel states are in use
          state_start = 1
          state_end = n_disc_state

          call proc%states_range(state_start, state_end)

          do indx_s = state_start, state_end

! ============ reference table of rules ============ !
!  1: clenshaw-curtis             2: clenshaw-curtis-zero
!  3: chebyshev                   4: chebyshev-odd
!  5: gauss-legendre              6: gauss-legendreodd
!  7: gauss-patterson             8: leja
!  9: lejaodd                    10: rleja
! 11: rleja-odd                  12: rleja-double-2
! 13: rleja-double-4             14: rleja-shifted
! 15: rleja-shifted-even         16: rleja-shifted-double
! 17: max-lebesgue               18: max-lebesgue-odd
! 19: min-lebesgue               20: min-lebesgue-odd
! 21: min-delta                  22: min-delta-odd
! 23: gauss-chebyshev-1          24: gauss-chebyshev-1-odd
! 25: gauss-chebyshev-2          26: gauss-chebyshev-2-odd
! 27: fejer2                     28: gauss-gegenbauer
! 29: gauss-gegenbauer-odd       30: gauss-jacobi
! 31: gauss-jacobi-odd           32: gauss-laguerre
! 33: gauss-laguerre-odd         34: gauss-hermite
! 35: gauss-hermite-odd          36: custom-tabulated
! 37: localp                     38: localp-zero
! 39: semi-localp                40: wavelet

! ============ reference table of grid types ============ !
!  1: level                       2: curved
!  3: iptotal                     4: ipcurved
!  5: qptotal                     6: qpcurved
!  7: hyperbolic                  8: iphyperbolic
!  9: qphyperbolic               10: tensor
! 11: iptensor                   12: qptensor

! ============ reference table of refinement types ============ !
!  1: classic            2: parents first
!  3: directional        4: FDS (both parents and directions)

! ============ reference table of acceleration types ============ !
!  0: none       1: CPU BLAS    2: GPU cuBLAS
!  3: GPU CUDA   4: GPU MAGMA


                if(step .eq. end_iter .or. policy_output) then
                     dim_out=nvars+1
                else
                     dim_out=1
                endif
                call tsgAllocateGrid(grid_new(indx_s))  !init object
                
                CALL tsgMakeLocalPolynomialGrid(grid_new(indx_s), dim_sim, dim_out, min_level, 1, tsg_localp)  !last two entries: function order, local grid type
                CALL tsgSetDomainTransform(grid_new(indx_s), statespace_down, statespace_up) !transform to acutal domain

                wtimeomp = MPI_Wtime( )

                do indx_ref=min_level,MAX(max_level,min_level) !we excedute at least one iteration independent of what max_level is

                        if(indx_ref > min_level) then
                             CALL tsgSetLocalSurplusRefinement(grid_new(indx_s),refinment_crit, tsg_fds)  !refine the grid if we are not in the base case
                        endif

		        points => tsgGetNeededPoints(grid_new(indx_s))
		        no_grid_pts_vec(indx_s) = tsgGetNumNeeded(grid_new(indx_s))

                        write(*,*) "Now solving",no_grid_pts_vec(indx_s),"points in refinment level",indx_ref

		        state_c = indx_s

!-----------------------------------------------------------------------

		        indx_start = 1
		        indx_end = no_grid_pts_vec(indx_s)

!.........................................................................

!.....allocate solution vector for VF (entry 1) and the controls (other entries)      
		        allocate(output_solver(nvars+1))     
		        allocate(return_status_vec_loc(no_grid_pts_vec(indx_s)))

!=======================================================================
!.....loop over design points where we want to solve the DP problem

		        return_status_vec_loc(:) = .false.


		        allocate(solve_data_loc(no_grid_pts_vec(indx_s)*(nvars+1)))  !fit only VF with GPR


		        solve_data_loc(:) = 0.0



		        tmp_bound =  no_grid_pts_vec(indx_s)

		        !$OMP PARALLEL DO schedule(dynamic) PRIVATE(output_solver,j), &
		        !$OMP PRIVATE(grid_pt_box,indxd)
		        do indxi = indx_start, indx_end

!----------------------------------------------------------------------   
!.....get coords of an individual grid points.

		                grid_pt_box(:) = points(:,indxi)


!----------------------------------------------------------------------
!.....solve optimization problem 
		                if (which_optimizer .eq. 1) then
		                       if(step .eq. 1) then
	   		                      call nonlinear_solver(indx_s, grid_pt_box,&
	   		                      & output_solver(:), return_status_vec_loc(indxi), nvars+1)
		                       else
	   		                      call nonlinear_solver_iterate(indx_s, grid_pt_box, &
	   		                      & output_solver(:), return_status_vec_loc(indxi), nvars+1)
		                       endif
		                else if (which_optimizer .eq. 2) then 
		                       call nonlinear_solver_snopt(step, indx_s, grid_pt_box,&
		                       & output_solver(:), return_status_vec_loc(indxi), nvars+1)

		                endif

		                !write output to vectors

		                do indxd = 1,nvars+1
		                         solve_data_loc((indxi-1)*(nvars+1) + indxd) = output_solver(indxd)  !store VF and policies                          			                  
		                enddo


!.....write solution to nonlinear system of equations to file
		                if(debug .ge. 2 .and. n_process_comm .eq. 1) then

		                       write(2,*) "point no" , indxi, grid_pt_box(:) , "Variables", &
		                       &   (j, output_solver(j), j=1,DofPerNode)  !output VF and controls

		                end if



		        end do 
		        !$OMP END PARALLEL DO  



10		        if(step .eq. end_iter .or. policy_output ) then
		               allocate(solve_data(nvars+1,no_grid_pts_vec(indx_s)))  !fit VF and policies
		               do indxd = 1,nvars+1
		                      do indxi = indx_start, indx_end
		                            solve_data(indxd,indxi) = solve_data_loc((indxi-1)*(nvars+1) + indxd)  !store for tasmanian
		                      enddo
		               enddo
		        else
		               allocate(solve_data(1,no_grid_pts_vec(indx_s)))  !fit only VF

		               do indxi = indx_start, indx_end
		                   solve_data(1,indxi) = solve_data_loc((indxi-1)*(nvars+1) + 1)  !store for tasmanian
		               enddo

		        endif


		        allocate(return_status_vec(no_grid_pts_vec(indx_s)))

		        return_status_vec(:) = return_status_vec_loc(:)

		        !count feasible points and performe safety check
		        no_feas = 0
		        do indx_pt = 1,no_grid_pts_vec(indx_s)
		                 if(return_status_vec(indx_pt)) then
		                     no_feas = no_feas + 1
		                 endif
		        enddo
		        if(no_feas < no_grid_pts_vec(indx_s) .and. no_feas > 0.1*no_grid_pts_vec(indx_s)) then
		                   write(*,*) "WARNING: only found", no_feas, " feasible points out of ", &
		                   &                                            no_grid_pts_vec(indx_s),"points"
		        else if (no_feas .le. 0.1*no_grid_pts_vec(indx_s)) then 
		                   write(*,*) "ERROR: only found", no_feas, & 
		                   &          "points. More than 90% of points did not converge"
		                   stop
		        endif

!-----------------------------------------------------------------------        
!.......create new grid

		        CALL tsgLoadNeededPoints(grid_new(indx_s), solve_data)

!-----------------------------------------------------------------------        
!.......write to file

                        if(indx_ref .eq. MAX(max_level,min_level)) then
				write(format_string_step,'(I0)') step  !previous iterations' data
				write(format_string_st,'(I0)')  indx_s
				write(filename_type1, '(a,a)') trim(restart_dir)// &
				   &             'VF_step_'//trim(format_string_step)&
				   &                      //'_'//trim(format_string_st)
				if(proc%isMaster()) then
				        write(*,*)'writing results to ',trim(filename_type1)
				endif
				call tsgWrite(grid_new(indx_s), trim(filename_type1) // c_null_char, binaryio)
                        endif

!-----------------------------------------------------------------------        
!.......deallocate memory

                        if(indx_ref .eq. MAX(max_level,min_level)) then
		             CALL tsgDeallocateGrid(grid_new(indx_s)) !delete new grid only in final iteration
                        endif

		        deallocate(solve_data)

		        deallocate(output_solver)
		        deallocate(return_status_vec)

		        deallocate(return_status_vec_loc,solve_data_loc)

                	if (proc%isMaster() .and. indx_ref .eq. MAX(max_level,min_level) ) then
		                   write(*,*) "================================================================================="
		                   write(*,*) "Wall time for computing the Bellman equation in state",indx_s, "is", MPI_Wtime( )-wtimeomp
		                   write(*,*) "================================================================================="
		        endif


!=====================================================================
          	enddo !do indx_ref
!=====================================================================
          enddo !end state
!=====================================================================


          if(parstates) then
              call mpi_barrier(MPI_COMM_WORLD, mpi_ierr)
              call proc%MPI_Err_Check(mpi_ierr)
          else 
              call mpi_barrier(proc%getComm(), mpi_ierr)
              call proc%MPI_Err_Check(mpi_ierr)
          endif

!----------------------------------------------------------------------

          if(debug .ge. 2) then
                  close(unit=2)  !close output file      
          end if   

          do indx_s =1, n_disc_state

              call tsgDeallocateGrid(grid_old(indx_s))

          enddo

          if(proc%isMaster(MPI_COMM_WORLD)) then
                write(*,*) "================================================================================="  
                write(*,*) "Wall time in iteration", step , &
                         & "is",  MPI_Wtime()-wtime
                write(*,*) "================================================================================="  
          endif

!=====================================================================
      end do !end DP iterations      
!=====================================================================    
   
     
      
!=======================================================================

      end subroutine main_fortran

!*********************************************************************** 
      
