!=======================================================================
!
!     this subroutine sets the parameters and callback functions 
!     for IPOPT.
!
!     Problem we aim to solve is fernandes pheland
!
!     v1: Simon Scheidegger, 11/15
!     v2: Simon Scheidegger, 06/18
!     v3: Philipp Renner, 07/18
!=======================================================================

      module SNOPT_wrapper_sim_init


      use iso_c_binding      

      use parameter_module
      use econ_module
      use, intrinsic :: iso_fortran_env, only : stdin=>input_unit, &
                                          stdout=>output_unit, &
                                          stderr=>error_unit

      implicit none
    

   
      contains        


!=======================================================================
!
!     value function objective during VFI
!
!=======================================================================
      subroutine EV_F_SIM_INIT( Status, n, x, &
     &     needF, nF, F, &
     &     needG, lenG, G, &
     &     cu, lencu, iu, leniu, ru, lenru )

      implicit none

      integer ::  Status, needF, needG, nF, n, lenG, &
     &     lencu, leniu, lenru, &
     &     iu(leniu)
      real(8) ::   F(nF), G(lenG), x(n), &
     &     ru(lenru)
      character(LEN=8) ::  cu(lencu)

      integer ::  Obj  !location of objective function
 
      real(prec) :: VT_sum 
      real(prec) :: x_tmp(n)
   
      !finite difference variables         
      real(prec)    :: xhm(n),xhp(n)
      real(prec) :: h
      real(prec) :: fvalhm,fvalhp
      integer :: indx,indxh
 
!----------------------------------------------------------------------   
  
      Obj    = 1                ! index of objective row    
 

      x_tmp(:) = x(:)
      if (needF .gt. 0) then   !provide objective and constraints
!.....call current functions
 
              call objective_sim_init( n, x_tmp, &
              & VT_sum, iu, leniu, ru, lenru)

!.....output      
	      F(Obj) = VT_sum
  
      endif           

      if (needG .gt. 0) then

!.........computing derivative with finite difference
          h = 1d-4
          do indx =1,N
              xhm(:) = x_tmp(:)
              xhp(:) = x_tmp(:)
              indxh = indx
              xhm(indxh) = x(indxh) - h
              
              call objective_sim_init( n, xhm, &
              & fvalhm, iu, leniu, ru, lenru)

              xhp(indxh) = x(indxh) + h
              call objective_sim_init( n, xhp, &
              & fvalhp, iu, leniu, ru, lenru)
              G(indx) = (fvalhp - fvalhm)/(2.0*h)
          enddo


      endif
   
      return
!-----------------------------------------------------------------------      
      end subroutine EV_F_SIM_INIT
!-----------------------------------------------------------------------        

!-----------------------------------------------------------------------      
      subroutine  objective_sim_init(n,x,fval,iu,leniu, ru, lenru)
!-----------------------------------------------------------------------      
      implicit none

      integer :: n,lenru
      real(prec) :: x(n),fval,x_tmp(dim_sim),ru(lenru)
      integer :: leniu, iu(leniu)

            
      real(prec) :: V_inter
      real(prec) :: util
            
                        
!.....Fortran - C interface 
      integer :: nst, which_GP, which_state

      double precision :: query_pt(dim_sim),res(nvars+1)



!.....compute next periods utility by interpolating on GP

      V_inter = 0.

      which_state = iu(1)
      x_tmp(1:n) = x(:)


      nst = which_state      !which state are we in 
      which_GP = 0 !-> VF (starts at 0 - dim_out -1) used for multidimensional problem
      query_pt(:) = x_tmp(:)
      CALL tsgEvaluate(grid_old(which_state), query_pt, res)

      V_inter = res(1)

!.....compute value function      
	      
      fval = V_inter

!-----------------------------------------------------------------------      
      end subroutine objective_sim_init
!-----------------------------------------------------------------------    


     
      
!-----------------------------------------------------------------------   
      
      end module SNOPT_wrapper_sim_init
      
!***********************************************************************


