      subroutine main_fortran(n_error_pts,start_error_it,end_error_it)       
!***********************************************************************
!
!     error computation routine
!
!***********************************************************************

      use parameter_module 
      USE TasmanianSG
      USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_PTR, C_DOUBLE, C_INT, C_CHAR, c_null_char
      
      implicit none

!......Fortran - C interface 

      integer(C_INT), intent(in) :: end_error_it,start_error_it,n_error_pts 

      real(prec)                                ::  diff_vec(n_error_pts)
      real(prec)                                ::  diff_sq_vec(n_error_pts)
      real(prec), allocatable, dimension(:,:)   ::  error_L1,error_MAX,error_L2
      real(prec)   ::  error_L1_cum,error_MAX_cum,error_L2_cum
      
      integer :: nst      !which discrete state/type to look at

      double precision                         :: sample_pts(n_error_pts,dim_model)
      double precision                         :: x_probe(dim_model)
      integer(C_INT) ::  dim_out_prev,dim_out_next
      double precision                         :: res(nvars+1)
      

      type(TasmanianSparseGrid) :: handle_prev(n_disc_state)
      type(TasmanianSparseGrid) :: handle_next(n_disc_state)

      integer :: i,j,k,indxp,indxit


      integer, dimension(33) :: seed
      real(prec) :: rand_num

!.....location for restart files
      character(256) :: data_dir = '../../compile/restart/'      
      character(len=256) :: filename_type_prev
      character(len=256) :: filename_type_next
      character(len=256) :: filename_last
      character(len=256) :: filename,folder
      character(len=256) :: format_string_step,format_string_st,format_string_start,format_string_end    
      character(len=256) :: WT_restart, dim_subspace_restart      
        
!......................................................................
!.....basic info on model..............................................
      write(*,*) '------------------------------------'
      write(*,*) dim_sim, "-- dimension of problem"   
      write(*,*) nvars ,  "-- nvars -- number of controls"            

!.........................................................................
!.....generate random sample points for value function iteration error testing

      do j=1,n_error_pts
           seed(:) = (j*1234+4356)
           call random_seed (PUT = seed) 
           do i = 1,dim_model
               call random_number(rand_num)
               sample_pts(j,i) = statespace_down(i) + (statespace_up(i)-statespace_down(i))*rand_num
           enddo
      enddo

!......................................................................

 
      allocate(error_L1(end_error_it-start_error_it,n_disc_state),  & 
                &   error_MAX(end_error_it-start_error_it,n_disc_state), &
                &   error_L2(end_error_it-start_error_it,n_disc_state))

             


      do indxit = start_error_it+1,end_error_it


!......................................................................
!.........generate filename     
	      do nst=1,n_disc_state
!.......................load previous itertion
			write(format_string_step,'(I0)') indxit-1  !previous iterations' data
			write(format_string_st,'(I0)')  nst
			write(filename_type_prev, '(a,a)') trim(data_dir)//'VF_step_'//trim(format_string_step)&
		      &                      //'_'//trim(format_string_st)
			write(*,*)'fn=',trim(filename_type_prev)
                        call tsgAllocateGrid(handle_prev(nst))  !init object
                        if (.not. tsgRead(handle_prev(nst), trim(filename_type_prev) // c_null_char)) then 
                           write(*,*) "File read failed: 1"
                           stop 1
                        end if
!.......................load current itertion
			write(format_string_step,'(I0)') indxit  !current iterations' data

			write(filename_type_next, '(a,a)') trim(data_dir)//'VF_step_'//trim(format_string_step)&
		      &                      //'_'//trim(format_string_st)
			write(*,*)'fn=',trim(filename_type_next)
                        call tsgAllocateGrid(handle_next(nst))  !init object
                        if (.not. tsgRead(handle_next(nst), trim(filename_type_next) // c_null_char)) then 
                           write(*,*) "File read failed: 1"
                           stop 1
                        end if


                        diff_vec(:) = 0.
                        diff_sq_vec(:) = 0.
                        !$OMP PARALLEL DO schedule(dynamic) private(x_probe, res)
                        do indxp=1,n_error_pts	   
!......................................................................
!.....call interpolation 
  
			      x_probe(:) = sample_pts(indxp,:)

	      
!.....interpolate      
                       	      CALL tsgEvaluate(handle_next(nst), x_probe, res)
                              diff_vec(indxp) = res(1)

                              CALL tsgEvaluate(handle_prev(nst), x_probe, res)
                              diff_vec(indxp) = ABS(diff_vec(indxp)-(res(1)))
                              diff_sq_vec(indxp) = diff_vec(indxp)**2
		      
                        enddo  !do indxp=1,n_error_pts
                        !$OMP END PARALLEL DO                     

                        error_MAX(indxit-start_error_it,nst) = MAXVAL(diff_vec)
                        error_L1(indxit-start_error_it,nst) = SUM(diff_vec)/n_error_pts
                        error_L2(indxit-start_error_it,nst) = SQRT(SUM(diff_sq_vec))/n_error_pts
	      
	      end do !end nst=1,n_disc_state  

!......................................................................
!.....Release memory of GPs
              do nst=1,n_disc_state 
   	        CALL tsgDeallocateGrid(handle_prev(nst))
	        CALL tsgDeallocateGrid(handle_next(nst))  
              enddo

      enddo  !end indxit


!.......................................................................
!.....Write errors to file
      do j=1,n_disc_state
		  write(format_string_start,'(I0)')  start_error_it
		  write(format_string_end,'(I0)')  end_error_it
   	          write(format_string_st,'(I0)')  j
	          write(filename,'(a)') 'error_'//trim(format_string_start)//'_'// & 
                       &   trim(format_string_end)//'_'//trim(format_string_st)//'.txt'
	          write(*,*)'fn=',trim(filename)      
	          open (unit = 2, file = filename)  
                  write(2,*) "#Errors iteration L1 L2 MAX"
                  do i= start_error_it+1,end_error_it
                      write(2,*) i , error_L1(i-start_error_it,j), & 
                       &  error_L2(i-start_error_it,j), error_MAX(i-start_error_it,j)
                  enddo !i=start
      enddo !j=1,n_disc_state

      write(format_string_start,'(I0)')  start_error_it
      write(format_string_end,'(I0)')  end_error_it
      write(filename,'(a)') 'error_'//trim(format_string_start)//'_'// & 
      &   trim(format_string_end)//'.txt'
      write(*,*)'fn=',trim(filename)      
      open (unit = 2, file = filename)  
      write(2,*) "#Errors iteration L1 L2 MAX"
      do i= start_error_it+1,end_error_it
          error_L1_cum=0.
          error_L2_cum=0.
          error_MAX_cum=0.
          do j = 1,n_disc_state
             error_L1_cum=error_L1_cum+error_L1(i-start_error_it,j)
             error_L2_cum=error_L2_cum+error_L2(i-start_error_it,j)
             error_MAX_cum=error_MAX_cum+error_MAX(i-start_error_it,j)
          enddo
          error_L1_cum=error_L1_cum/n_disc_state
          error_L2_cum=error_L2_cum/n_disc_state
          error_MAX_cum=error_MAX_cum/n_disc_state
          write(2,*) i , error_L1_cum, & 
              &  error_L2_cum,  error_MAX_cum
      enddo !i=start

!......................................................................
      write(6,*) '------------------------------------'      
      write(6,*) 'Postprocessing done!'      

      deallocate(error_L1, error_MAX, error_L2)
      
!======================================================================

      end subroutine main_fortran

!********************************************************************** 
      
