!=======================================================================
!
!     subroutine nonlinear_solver_sim 
!
!
!=======================================================================


      
      subroutine nonlinear_solver_sim(which_state, grid_pt_box, &
                       & start_pt, use_warmstart, output_solver, return_status)
  

      use, intrinsic :: iso_fortran_env, only : stdin=>input_unit, &
                                          stdout=>output_unit, &
                                          stderr=>error_unit

      USE, INTRINSIC :: ISO_C_BINDING
      

      use econ_module

      use IPOPT_wrapper, only: EV_F_ITER,EV_GRAD_F_ITER, EV_HESS,EV_G,EV_JAC_G 
                  
      use parameter_module
      
  
      implicit none
      
      include 'IpReturnCodes_SimonS.inc'  !include IPOPT            
      
!.....input variables      
      integer, intent(in) :: which_state!in which state are we 
      integer, intent(in) :: use_warmstart!in which state are we 
      real(prec), dimension(nvars), intent(in) :: start_pt  !in which state are we 
      real(prec), dimension(dim_sim), intent(in) :: grid_pt_box !coords one single grid point on a cubic box

!----------------------------------------------------------------------      
!     IPOPT PARAMETERS below:
!----------------------------------------------------------------------
!.....Size of the problem (number of variables and equality constraints)       
      integer :: N != dim_sim + 1, !number of vars.
      integer :: M != dim_sim + 1, !number of constraints
      integer :: NELE_JAC != M*N, !entries of Jacobian matrix of contraints
      integer :: NELE_HESS != (N**2 - N)/2 + N, !number of non-zero entries of Hess matrix            
      integer, parameter :: IDX_STY = 1  !index style = fortran style

!.....objective function for IPOPT      
      real(prec) :: F  !objective function      
      
!.....Space for multipliers and constraints
      real(prec), allocatable ::  LAM(:)  !multipliers
      real(prec), allocatable ::  G(:)    !(in-)equality constraints
      
!.....Vector of variables -> solution of non-linear equation system 
      real(prec)  ::  X(nvars)      
      
!.....Vector of lower and upper bounds on multipliers
      real(prec), allocatable ::  Z_L(:), Z_U(:)
      
!.....Private data for evaluation routines
!     This could be used to pass real(prec) ::  and integer arrays untouched
!     to the evaluation subroutines EVAL_*
      real(prec) :: DAT(dim_sim)
      integer    :: IDAT(1)

!.....Place for storing the Ipopt Problem Handle
!.....for 64 bit platforms:
      integer(8) :: IPROBLEM
      integer(8) :: IPCREATE
     
      integer :: NEW_X
      integer :: IERR
      integer :: IPSOLVE, IPADDSTROPTION
      integer :: IPADDNUMOPTION, IPADDINTOPTION
      integer :: IPOPENOUTPUTFILE

!-----------------------------------------------------------------------        
!.....local variables                
!-----------------------------------------------------------------------  
 
      real(prec), dimension(nrestarts,nvars)  :: X_start 
      integer :: i,j,max_indx  !index
      real(prec) :: rand_num, pen_val(n_disc_state)
      real(prec), allocatable ::  results_allrestarts(:,:)  !stores all policy values, objective value and return status for each restart
      integer, dimension(33) :: seed
      real(prec), dimension(dim_sim) :: grid_pt !coords one single grid point grid_pt(dim) in [0..1]^d


!.....state space bounds
      real(C_DOUBLE), save, dimension(dim_model) :: state_up
      real(C_DOUBLE), save, dimension(dim_model) :: state_down

!-----------------------------------------------------------------------        
!.....debug variables                
!-----------------------------------------------------------------------  

!      real(prec), dimension(:,:), allocatable   :: test_mat
!      real(prec), dimension(:), allocatable     :: test_likel_vec
!      real(prec), dimension(dim_sim)          :: temp_vec
!      integer                                   :: no_pts

!-----------------------------------------------------------------------        
!.....output variables                
!----------------------------------------------------------------------- 
      real(prec)                              :: VF                      !default case
      
      real(prec), intent(out) :: output_solver(2+nvars+dim_model)
      integer, intent(out) :: return_status !0 if converged -1 otherwise
              
!==================================================================
!==== Main function ==============================================
!==================================================================




!.....define problem size and allocate corresponding memory
      N = nvars    !number of vars.     
      M = nconstr    !number of constraints
      NELE_JAC = no_nz_jac !28     !entries of Jacobian matrix of contraints
      NELE_HESS = (N**2 - N)/2 + N !number of non-zero entries of Hess matrix       
     

      
      allocate(LAM(M), G(M))         !size M arrays 
      allocate(Z_L(N), Z_U(N) ) !size N arrays          
      allocate(results_allrestarts(nrestarts,N+3))   !stores all policy values, objective value and return status
      
      
        
!-------------------------------------------------------------------------
!......Open an output file
!...commented out in order to suppress some output problems
!      IERR = IPOPENOUTPUTFILE(IPROBLEM, 'IPOPT.OUT', 5)
!      if (IERR.ne.0 ) then
!         write(6,*) 'Error opening the Ipopt output file.'
!      endif      
!                    
!-----------------------------------------------------------------------      
!     First create a handle for the Ipopt problem (and read the options
!     file)
      IPROBLEM = IPCREATE(N, X_L, X_U, M, G_L, G_U, NELE_JAC, NELE_HESS, &
     &           IDX_STY, EV_F_ITER, EV_G, EV_GRAD_F_ITER, EV_JAC_G, EV_HESS)
       
!-----------------------------------------------------------------------      
!.....IPOPT OPTIONS
!.....Set a string option
      IERR = IPADDSTROPTION(IPROBLEM, 'mu_strategy', 'adaptive')

      IERR = IPADDSTROPTION(IPROBLEM, 'check_derivatives_for_naninf', 'no')
      IERR = IPADDSTROPTION(IPROBLEM, 'nlp_scaling_method', 'gradient-based')
      IERR = IPADDSTROPTION(IPROBLEM, 'derivative_test', 'none')
!.....Set the linear solver for IPOPT
      IERR = IPADDSTROPTION(IPROBLEM, 'linear_solver', 'ma27')
!.....Set an integer option
      IERR = IPADDINTOPTION(IPROBLEM, 'max_iter', 500)
!.....Set a real option
      IERR = IPADDNUMOPTION(IPROBLEM, 'tol', 1e-3)
        
      IERR = IPADDNUMOPTION(IPROBLEM, 'acceptable_tol', 1e-3)   
       
!.....switch off hessian -> check hessian fct. below
      IERR = IPADDSTROPTION(IPROBLEM, 'hessian_approximation',  &
     &                                'limited-memory')
!.....print user options
      IERR = IPADDSTROPTION(IPROBLEM, 'print_user_options', 'no') 
         
!.....restauration phase, desired threshold for constraint violation
      IERR = IPADDNUMOPTION(IPROBLEM, 'constr_viol_tol',1e-3)      
        
      IERR = IPADDNUMOPTION(IPROBLEM, 'obj_scaling_factor', -1.0)    !maximization problem -> neg value      
         
         
!-----------------------------------------------------------------------       
!.....below: debug options for IPOPT !!!      
      IERR = IPADDSTROPTION(IPROBLEM, 'derivative_test', 'none') !leave away to gain speed
      IERR = IPADDSTROPTION(IPROBLEM, 'derivative_test_print_all', 'no') !leave away to gain speed      
     !IERR = IPADDNUMOPTION(IPROBLEM, 'derivative_test_perturbation', 1d-8)
!      ! olaf
      if(debug .le. 4) then
          IERR = IPADDINTOPTION(IPROBLEM, 'print_level', 0) !default 5, max 12
      else if(debug .ge. 5) then
          IERR = IPADDINTOPTION(IPROBLEM, 'print_level', 5) !default 5, max 12
      endif
! 
! 

!-----------------------------------------------------------------------       
!      passing info to IPOPT
 
      do j =1,dim_model
         DAT(j) = grid_pt_box(j)
      enddo

      call sparsity_pattern_obj(AVAR_OBJ, no_nz_obj)
      IDAT(1) = which_state
      IDAT(2:1+no_nz_obj) = AVAR_OBJ(:)

!...............................................................................
!.....bounds on constraints and variables

      call bounds(n,X_L, X_U, m, G_L, G_U, bigbound, &
      &                                          IDAT, DAT )

!...............................................................................
!.....feasible starting points


      do j=1,nrestarts
           seed(:) = (j*1234+4356)
           call random_seed (PUT = seed) 
           do i = 1,nvars
               call random_number(rand_num)
               X_start(j,i) = X_L(i) + (X_U(i)-X_L(i))*rand_num
           enddo
           call solve_feas(grid_pt_box, X_start(j,i), nvars, DAT, IDAT) !giving a better guess than random
      enddo


      do i =1,nrestarts
!.....initial guess
!         seed(:) = (i*1234+4356)
!         call random_seed (PUT = seed) 
!         do j = 1,nvars 
!            call random_number(rand_num)
!            X(i,:) = X_L(j) + (X_U(j) - X_L(j))*rand_num
!         enddo
         if (debug .ge. 5) then
            write(*,*) "==================================================="
            write(*,*) "Start point",i,"for point",grid_pt_box,"is",X_start(i,:)
            write(*,*) "==================================================="
         endif
         IERR = IPSOLVE(IPROBLEM, X_start(i,:), G, F, LAM, Z_L, Z_U, IDAT, DAT)
         X(:) = X_start(i,:)
         if(IERR .eq. IP_SOLVE_SUCCEEDED .or. IERR .eq. IP_ACCEPTABLE_LEVEL) then

             results_allrestarts(i,1) = 0
             results_allrestarts(i,2) = F
             if(debug .ge. 3) then
                 write(*,*) "==================================================="
                 write(*,*) "Intermediate result in nonlinear_solver_iterate"
                 write(*,*) "Current point", grid_pt_box(:)
                 write(*,*) "Start point no", i
                 write(*,*) "Objective",results_allrestarts(i,2) 
                 write(*,*) "Optimal solution",X
                 write(*,*) "==================================================="
             endif
         else
             results_allrestarts(i,1) = -1
             results_allrestarts(i,2) = -1d20
         endif
           
         results_allrestarts(i,4:N+3) = X(:)
      enddo


!-----------------------------------------------------------------------            
!.....Output & refinement criterion:

      output_solver(1) = -1d20 !objective
      output_solver(2) = -1d20 !penalty value
      output_solver(3:) = 0. !optimal solution
      return_status = 0
      do i =1,nrestarts

          if(  results_allrestarts(i,1) .eq. 0 .and. & 
             & results_allrestarts(i,2) .ge. output_solver(1)) then
              
              output_solver(1) = results_allrestarts(i,2)
              output_solver(2) = results_allrestarts(i,3)
              output_solver(3:nvars+2) = results_allrestarts(i,4:nvars+3)
              return_status = 1

          endif

      enddo
      if( return_status .eq. 1 ) then
            if(debug .ge. 2) then
                 write(*,*) "========================================="
                 write(*,*) "CONVERGED in nonlinear_solver_iterate"
                 write(*,*) "Current point", grid_pt_box(:)
                 write(*,*) "Objective",output_solver(1) 
                 write(*,*) "========================================="
            endif
      elseif(MAXVAL(results_allrestarts(:,1)) .eq. 0) then

            if(debug .ge. 2) then
                 write(*,*) "=============================================================="
                 write(*,*) "FAILED TO CONVERGE DUE TO PENALTY in nonlinear_solver_iterate"
                 write(*,*) "Current point", grid_pt_box(:)
                 write(*,*) "Objective",output_solver(1) 
                 write(*,*) "Penalty", results_allrestarts(max_indx,3)
                 write(*,*) "==============================================================="
            endif
      else

            if(debug .ge. 2) then        
               write(*,*) "============================================================"
               write(*,*) 'Nonlinear_solver_iterate did not converge at any point.'
               write(*,*) "Current point", grid_pt_box(:)
               write(*,*) "============================================================"
            endif           

      endif
        

! !.....Clean up
     call IPFREE(IPROBLEM)
            
!-----------------------------------------------------------------------      
     deallocate(LAM, G )         !size M arrays 
     deallocate(Z_L, Z_U)    !size N arrays     
     deallocate(results_allrestarts) 
      
!-----------------------------------------------------------------------        
     
     
      
      

!=======================================================================

     end subroutine nonlinear_solver_sim

!***********************************************************************
      
      
