!=======================================================================
!
!     subroutine nonlinear_solver_sim_snopt_init 
!
!
!=======================================================================


      
      subroutine nonlinear_solver_sim_snopt_init(which_state, &
                               &  output_solver, return_status)
  

      use, intrinsic :: iso_fortran_env, only : stdin=>input_unit, &
                                          stdout=>output_unit, &
                                          stderr=>error_unit

      USE, INTRINSIC :: ISO_C_BINDING
      

      use SNOPT_wrapper_sim_init, only: EV_F_SIM_INIT 
                  
      use parameter_module
      use econ_module
      
  
      implicit none
      
      include 'IpReturnCodes_SimonS.inc'  !include IPOPT            
      
!.....input variables      
      integer, intent(in)    :: which_state!in which state are we 

!----------------------------------------------------------------------      
!     IPOPT PARAMETERS below:
!----------------------------------------------------------------------
!.....Size of the problem (number of variables and equality constraints)       
      integer :: M != dim_sim + 1, !number of constraints
       
!-----------------------------------------------------------------------        
!.....SNOPT variables
!-----------------------------------------------------------------------  

      integer,  parameter :: maxF   = 1, maxn = dim_sim, &
     &     lenA   = 1, &
     &     lenG   = dim_sim, &
     &     nxname =  1, nFname =   1 
      integer :: iAfun(lenA), jAvar(lenA), iGfun(lenG), jGvar(lenG), &
     &     xstate(maxn), Fstate(maxF)
      CHARACTER(LEN=8) ::  ProbN(10), xnames(nxname), Fnames(nFname)
      real(8) :: ObjAdd, sInf,&
     &     A(lenA), Flow(maxF), Fupp(maxF), F(maxF), &
     &     xlow(maxn), xupp(maxn), Fmul(maxF), xmul(maxn)
      integer ::  DerOpt, neA, neG, ObjRow, INFO, iPrt, iPrint, iSpecs, &
     &     iSum, iSumm, Mjrlim, mincw, miniw, minrw, nF, n, nInf, &
     &     nOut, nS, Obj

      integer :: lunit
      CHARACTER(LEN=20) :: lfile

!     ------------------------------------------------------------------
!     SNOPT workspace

      integer, parameter :: lenrw = 800*(dim_sim + 1)
      real(8) ::  rw(lenrw)
      integer,  parameter :: leniw = 200*(dim_sim + 1)
      integer  ::       iw(leniw)
      integer , parameter :: lencw =   500 + dim_sim + 1
      CHARACTER(LEN=8)    ::  cw(lencw)

!     ------------------------------------------------------------------
!     SNOPT variables used for passing information

      integer ,parameter :: lenru = 1
      real(8)  :: ru(lenru)
      integer,parameter ::  leniu = 2
      integer  ::   iu(leniu)
      integer   ,parameter ::  lencu = 1 
      character(LEN=8)  ::  cu(lencu)

!     ------------------------------------------------------------------
      integer ,parameter :: Cold   = 0, Basis  = 1, Warm  = 2
!     ------------------------------------------------------------------


!-----------------------------------------------------------------------        
!.....local variables                
!-----------------------------------------------------------------------  
 
      real(C_DOUBLE), dimension(nrestarts,dim_sim)  :: X_start 
      double precision, dimension(dim_sim)  :: X
      integer :: i,j,max_indx  !index
      real(prec) :: rand_num, pen_val
      real(prec), allocatable ::  results_allrestarts(:,:)  !stores all policy values, objective value and return status for each restart
      integer, dimension(33) :: seed


      real(C_DOUBLE), save, dimension(dim_sim) :: grid_pt,grid_pt_box

!-----------------------------------------------------------------------        
!.....debug variables                
!-----------------------------------------------------------------------  

!      real(prec), dimension(:,:), allocatable   :: test_mat
!      real(prec), dimension(:), allocatable     :: test_likel_vec
!      real(prec), dimension(dim_sim)          :: temp_vec
!      integer                                   :: no_pts

!-----------------------------------------------------------------------        
!.....output variables                
!----------------------------------------------------------------------- 
      
      real(prec), intent(out) :: output_solver(2+dim_sim)
      integer, intent(out) :: return_status !0 if converged -1 otherwise
 
      external EV_F             
!==================================================================
!==== Main function ==============================================
!==================================================================


!.....define problem size and allocate corresponding memory
      N = dim_sim    
      M = 0    !number of constraints
      neG = lenG  !28     !entries of Jacobian matrix of contraints
      neA = 0   
      A(:) = 0 !no entries in linear constraints
      iw(:) = 0
      rw(:)=0.
      cw(:) = ""         
     
      allocate(results_allrestarts(nrestarts,N+3))   !stores all policy values, objective value and return status
      
!     ------------------------------------------------------------------
!     Specify some of the SNOPT files.
!     iSpecs  is the Specs file   (0 if none).
!     iPrint  is the Print file   (0 if none).
!     iSumm   is the Summary file (0 if none).
!
!     nOut    is an output file used here by hs47


      iSpecs = 0
      iSumm  = 0
      iPrint = 0


!        Unix and DOS systems.  Open the Specs and print files.

!         lunit = iSpecs
!         lfile = 'hsmaina.spc'
!         open( lunit, file=lfile, status='OLD',     err=800 )
!         lunit = iPrint
!         lfile = 'hsmaina.out'
!         open( lunit, file=lfile, status='UNKNOWN', err=800 )

!     ------------------------------------------------------------------
!     First,  snInit MUST be called to initialize optional parameters
!     to their default values.
!     ------------------------------------------------------------------

      iPrt   = iPrint
      iSum   = iSumm
      call snInit( iPrint, iSumm, cw, lencw, iw, leniw, rw, lenrw )



!     ------------------------------------------------------------------
!     Give the problem a name.

      ProbN   = 'P'  

      nF     = 1        ! constraints and objective
      Obj    = 1                ! index of objective row
      ObjRow = 1        ! Indx where objective is stored


     
      Flow(1) = -bigbound   !bounds on the objective
      Fupp(1) = bigbound



      xlow(:) = statespace_down(:)
      xupp(:) = statespace_up(:)

      xstate(:) =  0    !is a set of initial states for each x  (0,1,2,3,4,5).

      Fmul(:) = 0       !is a set of initial values for the dual variables.


      do j = 1,dim_sim  !derivative with respect to transfer
         iGfun(j) = ObjRow
         jGvar(j) = j
      enddo


      X_start(:,:) = 0.

      do j=1,nrestarts
           seed(:) = (j*1234+4356)
           call random_seed (PUT = seed) 
           do i = 1,dim_model
               call random_number(rand_num)
               X_start(j,i) = statespace_down(i) + (statespace_up(i)-statespace_down(i))*rand_num
           enddo
      enddo


!     ------------------------------------------------------------------
!     Warn snOptA that userf does not set the derivatives.
!     The parameters iPrt and iSum may refer to the Print and Summary
!     file respectively.  Setting them to 0 suppresses printing.
!     ------------------------------------------------------------------


      DerOpt = -1
      call snSeti( 'Derivative option', DerOpt, iPrt, iSum, INFO,&
     &    cw, lencw, iw, leniw, rw, lenrw )
!	0 Only a “cheap” test will be performed, requiring 2 calls to usrfun for snOptA, and 2
!	calls to funcon and 3 calls to funobj for snOptB.
!	1 Individual gradients will be checked (with a more reliable test). A key of the form
!	“OK” or “Bad?” indicates whether or not each component appears to be correct.
!	2 Individual columns of the problem Jacobian will be checked.
!	3 Options 2 and 1 will both occur (in that order).
!	−1 Derivative checking is disabled.

      MjrLim = 500
      call snSeti( 'Major Iterations', MjrLim, iPrt, iSum, INFO,&
     &      cw, lencw, iw, leniw, rw, lenrw )

      call snSet('Maximize' , iPrt, iSum, INFO,&
     &     cw, lencw, iw, leniw, rw, lenrw )

      call snSet('Nonderivative linesearch' , iPrt, iSum, INFO,&
     &     cw, lencw, iw, leniw, rw, lenrw )

      call snSetr('Minor feasibility tolerance',1e-6 , iPrt, iSum, INFO, &
      &      cw, lencw, iw, leniw, rw, lenrw )
      call snSetr('Major feasibility tolerance',1e-6 , iPrt, iSum, INFO, &
      &       cw, lencw, iw, leniw, rw, lenrw )
      call snSetr('Major optimality tolerance',1e-3 , iPrt, iSum, INFO, &
      &      cw, lencw, iw, leniw, rw, lenrw )
!      call snSetr('Infinite Bound size',bigbnd , iPrt, iSum, INFO, &
!      &      cw, lencw, iw, leniw, rw, lenrw )

!     ------------------------------------------------------------------
!     Go for it, using a Cold start (Start = 0).
!     ------------------------------------------------------------------

      ObjAdd = 0.  !constant added to objective

 !-----------------------------------------------------------------------       
!      passing info to IPOPT

      iu(1) = which_state
      ru(:) = 0.
! 


!.....solve nonlinear system of equations.
      !write(*,*) "hello from thread ", rank



      do i =1,nrestarts

         call snOptA( Cold, nF, n, nxname, nFname,&
         &     ObjAdd, ObjRow, ProbN, EV_F_SIM_INIT,&
         &     iAfun, jAvar, lenA, neA, A,&
         &     iGfun, jGvar, lenG, neG,&
         &     xlow, xupp, xnames, Flow, Fupp, Fnames,&
         &     X_start(i,:), xstate, xmul, F, Fstate, Fmul,&
         &     INFO, mincw, miniw, minrw,&
         &     nS, nInf, sInf,&
         &     cu, lencu, iu, leniu, ru, lenru,&
         &     cw, lencw, iw, leniw, rw, lenrw )

         if (INFO .eq. 82 .or. INFO .eq. 83 .or. INFO .eq. 84) then
             write(*,*) "ERROR: workspace for SNOPT too small, error code",info
             exit
         end if



         if(INFO .eq. 1 .or. INFO .eq. 2 .or. INFO .eq. 3) then

             results_allrestarts(i,1) = 0
             results_allrestarts(i,2) = F(ObjRow)
             if(debug .ge. 3) then
                 write(*,*) "==================================================="
                 write(*,*) "Intermediate result in nonlinear_solver_iterate"
                 write(*,*) "Objective",results_allrestarts(i,2) 
                 write(*,*) "Optimal solution",X
                 write(*,*) "==================================================="
             endif
         else
             results_allrestarts(i,1) = -1
             results_allrestarts(i,2) = -1d20
         endif
           
         results_allrestarts(i,4:N+3) = X_start(i,:)
      enddo


!-----------------------------------------------------------------------            
!.....Output & refinement criterion:

      output_solver(1) = -1d20 !objective
      output_solver(3:dim_sim+2) = 0. !optimal solution
      return_status = 0
      do i =1,nrestarts

          if(  results_allrestarts(i,1) .eq. 0 .and. & 
             & results_allrestarts(i,2) .ge. output_solver(1)) then
              
              output_solver(1) = results_allrestarts(i,2)
              output_solver(2) = results_allrestarts(i,3)
              output_solver(3:dim_sim+2) = results_allrestarts(i,4:N+3)
              return_status = 1

          endif

      enddo

      
      if( return_status .eq. 1 ) then
            if(debug .ge. 2) then
                 write(*,*) "========================================="
                 write(*,*) "CONVERGED in nonlinear_solver_iterate"
                 write(*,*) "Objective",output_solver(1) 
                 write(*,*) "========================================="
            endif
      elseif(MAXVAL(results_allrestarts(:,1)) .eq. 0) then

            if(debug .ge. 2) then
                 write(*,*) "=============================================================="
                 write(*,*) "FAILED TO CONVERGE DUE TO PENALTY in nonlinear_solver_iterate"
                 write(*,*) "Objective",output_solver(1) 
                 write(*,*) "==============================================================="
            endif
      else

            if(debug .ge. 2) then        
               write(*,*) "============================================================"
               write(*,*) 'Nonlinear_solver_iterate did not converge at any point.'
               write(*,*) "============================================================"
            endif           

      endif
        

! !.....Clean up


            
!-----------------------------------------------------------------------      
     deallocate(results_allrestarts) 
      
!-----------------------------------------------------------------------        
     

!=======================================================================

     end subroutine nonlinear_solver_sim_snopt_init

!***********************************************************************
      
      
