      subroutine plot_fortran(plotit,level)       
!***********************************************************************
!
!     main simulation routine
!
!     NOTE: Parameters need to be set in parameter_module.f90
!
!***********************************************************************

      use parameter_module
      USE TasmanianSG

      USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_PTR, C_DOUBLE, C_INT, C_CHAR, c_null_char
      
      implicit none
      include 'mpif.h'

!......Fortran - C interface 
      external :: limbo_load     

      integer(C_INT), intent(in) :: plotit,level
      
      integer :: nst      !which discrete state/type to look at

!.........................
      real(prec), dimension(2+nvars)        :: output_solver 
      integer :: return_status,which_state

      
!.....location for restart files
      character(256) :: data_dir = '../../compile/restart/'      
      character(len=256) :: filename_type_prev
      character(len=256) :: filename_type
      character(len=256) :: filename
      character(len=256) :: format_string_step,format_string_st   
      character(len=256) :: WT_restart, dim_subspace_restart      

!.....tasmanian vars
      double precision        :: query_pt(dim_sim)  
      real(prec) :: X(nvars)
      double precision, pointer :: points(:,:)
      double precision, allocatable :: res(:)
      integer(C_INT), dimension(n_disc_state) :: No_grid_pts_vec   !Number of grid points for VFI
      integer(C_INT) :: dim_output(n_disc_state), dout

!.....simulation variables
      integer :: fileid

!.....other local variables
      integer :: indxs,indxp, indxi, indxj, indxsum, indx_s



!.........................................................................



!.....basic info on model..............................................
          write(*,*) '======================================================='
          write(*,*) "dimension of problem" , dim_model
          write(*,*) "nvars -- number of controls" , nvars     
          write(*,*) '-------------------------------------------------------'
          write(*,*) '======================================================='
      

!......................................................................
!.........generate filename     

      write(format_string_step,'(I0)') plotit  !current iterations' data
      do nst=1,n_disc_state

!.......................load current itertion
                write(format_string_st,'(I0)')  nst
		write(filename_type, '(a,a)') trim(data_dir)//'VF_step_'//trim(format_string_step)&
		&                      //'_'//trim(format_string_st)
		write(*,*)'fn=',trim(filename_type)
                call tsgAllocateGrid(grid_old(nst))  !init object
                if (.not. tsgRead(grid_old(nst), trim(filename_type) // c_null_char)) then 
                      write(*,*) "File read failed: 1"
                      stop 1
                end if

                dim_output(nst) = MAX(2,tsgGetNumOutputs(grid_old(nst)))

                dout = 1
                call tsgAllocateGrid(grid_new(nst))  !init object
                CALL tsgMakeGlobalGrid(grid_new(nst), dim_sim, 0, level, tsg_level, tsg_clenshaw_curtis)  !last two entries: function order, local grid type
                CALL tsgSetDomainTransform(grid_new(nst), statespace_down, statespace_up) !transform to acutal domain

	      
      end do !end nst=1,n_disc_state  


!...............................................................................
!...............................................................................
!.........plot
      do nst=1,n_disc_state


        
          fileid = 666

          write(format_string_step,'(I0)')  plotit
          write(format_string_st,'(I0)')  nst
          write(filename,'(a)') 'VF_plot_'//trim(format_string_step)//'_'// & 
           &   trim(format_string_st)//'.txt'
          write(*,*)'fn=',trim(filename)      
          open(UNIT=fileid, FILE=filename )
          write(fileid,*) "#output sequence: value corrdinates variables"
          flush(fileid)

          points => tsgGetPoints(grid_new(nst))
          no_grid_pts_vec(nst) = tsgGetNumPoints(grid_new(nst))
          allocate(res(dim_output(nst)))
          res(:) = 0.

          do indxp=1,no_grid_pts_vec(nst)	   

     
!.......plot points

	              query_pt(:) = points(:,indxp)


                      CALL tsgEvaluate(grid_old(nst), query_pt, res)

		      
                      write(fileid,*) res(1), query_pt(:), &
                      &              res(2:dim_output(nst))
                      flush(fileid)

      
          enddo  !indxp=1,no_grid_pts_vec(nst)	

          
10        close(fileid)
          deallocate(res)

      enddo !do nst=1

!......................................................................
!.....Release memory of GPs

      do nst=1,n_disc_state
         CALL tsgDeallocateGrid(grid_new(nst))
         CALL tsgDeallocateGrid(grid_old(nst))
      enddo

!......................................................................
      write(6,*) '------------------------------------'      
      write(6,*) 'Plotting done!'      


      
!======================================================================

      end subroutine plot_fortran

!********************************************************************** 
      
