      subroutine sim_fortran(step,n_iter,optimize,use_warmstart,fix_shock)       
!***********************************************************************
!
!     main simulation routine
!
!     NOTE: Parameters need to be set in parameter_module.f90
!
!***********************************************************************

      use parameter_module
      use econ_module
      USE TasmanianSG

      USE, INTRINSIC :: ISO_C_BINDING, ONLY: C_PTR, C_DOUBLE, C_INT, C_CHAR, c_null_char
      
      implicit none
      include 'mpif.h'

!......Fortran - C interface 
      external :: limbo_load     

      integer(C_INT), intent(in) :: step   !what Gaussian process
      integer(C_INT), intent(in) :: n_iter !how many simulation iterations
      integer(C_INT), intent(in) :: optimize !0: no optimization during simulation; 1: optimization to help smooth policy
      integer(C_INT), intent(in) :: use_warmstart
      integer(C_INT), intent(in) :: fix_shock
      
      integer :: nst      !which discrete state/type to look at

!.....interplating vars      
      double precision        :: query_pt(dim_sim)  
      double precision        :: res(nvars+1)
      real(prec) :: X(nvars)

!.........................
      real(prec), dimension(2+nvars)        :: output_solver 
      integer :: return_status,which_state

      
!.....location for restart files
      character(256) :: data_dir = '../../compile/restart/'      
      character(len=256) :: filename_type_prev
      character(len=256) :: filename_type
      character(len=256) :: filename
      character(len=256) :: format_string_step,format_string_st   
      character(len=256) :: WT_restart, dim_subspace_restart      

!.....simulation variables
      integer :: shock_list(n_iter),fileid
      real(prec) shock_list_01(n_iter)
      integer :: stab
      real(prec) :: output_init(2+dim_model), state_in(dim_model)
      real(prec), dimension(no_integnodes,dim_sim) :: f_next_all
      integer, dimension(33) :: seed
      real(prec) :: rand_num, prob_cum
      real(prec)   ::  Prob(no_integnodes)

!.....other local variables
      integer :: indxs,indxp, dim_output, dout, indxi, indxj, indxsum, indx_s

      real(prec), dimension(nvars)          :: start_pt


      INTEGER, EXTERNAL :: OMP_GET_THREAD_NUM, OMP_GET_NUM_THREADS      

!.....MPI
      integer :: ierr,rank,mpisize



!........................................................................
!.....MPI setup

      call MPI_Comm_rank ( MPI_COMM_WORLD, rank, ierr )
      call MPI_Comm_size ( MPI_COMM_WORLD, mpisize, ierr ) 


!.........................................................................



!.....basic info on model..............................................
      if( proc%isMaster()) then
          write(*,*) '======================================================='
          write(*,*) "dimension of problem" , dim_model
          write(*,*) "nvars -- number of controls" , nvars     
          write(*,*) '-------------------------------------------------------'
          write(*,*) '======================================================='
      endif
      


      if(optimize .eq. 0) then 
         dout = nvars+1
      else
         dout = 1
      endif

!......................................................................
!.........generate filename     

      write(format_string_step,'(I0)') step  !current iterations' data
      do nst=1,n_disc_state

!.......................load current itertion
                write(format_string_st,'(I0)')  nst
		write(filename_type, '(a,a)') trim(data_dir)//'VF_step_'//trim(format_string_step)&
		&                      //'_'//trim(format_string_st)
		write(*,*)'fn=',trim(filename_type)
                call tsgAllocateGrid(grid_old(nst))  !init object
                if (.not. tsgRead(grid_old(nst), trim(filename_type) // c_null_char)) then 
                      write(*,*) "File read failed: 1"
                      stop 1
                end if
	      
      end do !end nst=1,n_disc_state  


!...............................................................................
!...............................................................................
!.........simulate
!      !$OMP PARALLEL DO schedule(dynamic) DEFAULT(FirstPrivate),shared(handle_GP), &
!      !$OMP SHARED(state_down,state_up,n_iter,optimize,step)
      do nst=1,n_disc_state

       if(MOD((nst-1),mpisize) == rank) then

        
          fileid = 666 +  OMP_GET_THREAD_NUM() + rank

          write(format_string_step,'(I0)')  step
          write(format_string_st,'(I0)')  nst
          write(filename,'(a)') 'simulate_'//trim(format_string_step)//'_'// & 
           &   trim(format_string_st)//'.txt'
          write(*,*)'fn=',trim(filename)      
          open(UNIT=fileid, FILE=filename )
          write(fileid,*) "#output sequence: value corrdinates variables"
          flush(fileid)

!........set shocks for simulation
          seed(:) = (nst*1234+4356)
          call random_seed (PUT = seed) 
          do indxp=1,n_iter
              call random_number(rand_num)
              shock_list_01(indxp) = rand_num
          enddo


          shock_list(1) = nst !initial shock
          if (fix_shock .eq. 1) then !use artifical shock list defined by user
               shock_list(:) = nst
          endif

          call nonlinear_solver_sim_snopt_init(shock_list(1), output_init, return_status)
          write(*,*) "Initial solution: Status",return_status, &
          &           "Obj",output_init(1),& 
          &           "Start pt",output_init(3:dim_model+2),&
          &           "state",nst
          state_in(1:dim_model) = output_init(3:dim_model+2)   


          do indxp=1,n_iter	   

     
!.......optimize to simulate 

	              query_pt(:) = state_in(:)

                      CALL tsgEvaluate(grid_old(shock_list(indxp)), query_pt, res)

                      if(optimize .eq. 0) then 
                          output_solver(1) = res(1)
                          output_solver(2) = 0.
                          output_solver(3:nvars+2) = res(2:nvars+1) 
                          X(:) = output_solver(3:nvars+2)
                          return_status = 1

                      endif
                      if(optimize .eq. 1) then 
                          if(use_warmstart .eq. 1) then

                              start_pt(:) = res(2:nvars+1) 

                          else

                              start_pt(:) = 0.

                          endif

		          !call nonlinear_solver_sim(shock_list(indxp), state_in, &
                          !               & start_pt,use_warmstart, output_solver, return_status)
		          call nonlinear_solver_sim_snopt(shock_list(indxp), state_in, &
                                         & start_pt, use_warmstart, output_solver, return_status)
                          X(:) = output_solver(3:nvars+2)
                      endif

                      if(return_status .ne. 1) then
                         write(*,*) "====================================================="
                         write(*,*) "Failed to converged in state",nst,"iteration",indxp
                         write(*,*) "====================================================="
                         goto 10
                      endif

                      
                      if(optimize .eq. 1) then
                            write(*,*) "Iteration",indxp,"State",state_in(:),"shock",shock_list(indxp) &
                            &     , "error",abs(output_solver(1)-res(1)) 
		      
                            write(fileid,*) output_solver(1), state_in(:), shock_list(indxp), &
                             &              X(:)
                            flush(fileid)
                      else
                            write(*,*) "Iteration",indxp,"State",state_in(:),"shock",shock_list(indxp) &
                            &     , "error",abs(output_solver(1)-res(1)) 
		      
                            write(fileid,*) output_solver(1), state_in(:), shock_list(indxp), &
                             &           X(:)
                            flush(fileid)

                      endif

!.....................next state in old coordinates
                      call next_cont_state(f_next_all, shock_list(indxp), X(:), nvars)




!.....................determining the next cont and disc state                   
                      if(indxp < n_iter .and. fix_shock .eq. 0 .and. indxp .ge. 1) then
                             call prob_weights(X,shock_list(indxp),Prob)  !get prob weights
                             prob_cum = Prob(1) !Probability cumulative distribution
                             do indxi=1,no_integnodes
                                if(shock_list_01(indxp) < prob_cum) then
                                    shock_list(indxp+1) = indxi
                                    state_in(:) = f_next_all(indxi,:)
                                    goto 20
                                endif
                                if(indxi < no_integnodes) then
                                   prob_cum = prob_cum + Prob(indxi+1)
                                endif
                             enddo
                      else if(indxp < n_iter .and. fix_shock .eq. 1) then

!                                    shock_list(indxp+1) = nst
                                    state_in(:) = f_next_all(shock_list(indxp+1),:)
                      else
                                    shock_list(indxp+1) = nst
                                    state_in(:) = f_next_all(shock_list(indxp+1),:)

20                    endif

      
          enddo  !do indxp=1,n_error_pts


10        close(fileid)
       endif

      enddo !do nst=1
!      !$OMP END PARALLEL DO   
      call MPI_Barrier(MPI_COMM_WORLD,ierr)

!......................................................................
!.....Release memory of GPs

      do nst=1,n_disc_state
         CALL tsgDeallocateGrid(grid_old(nst))
      enddo

!......................................................................
      write(6,*) '------------------------------------'      
      write(6,*) 'Simulation done!'      


      
!======================================================================

      end subroutine sim_fortran

!********************************************************************** 
      
